%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#define __STDC__
#endif

int yylex();
int yyparse();

int existserr = 0;

struct envdef{
	float scrx, scry, scrz;
	float seedeg1, seedeg2;
	int lightnum;
	float lightdeg1, lightdeg2;
	float lightpow;
};

struct objectdef{
	int texture;
	int form;
	int reftype;
	int existrot;
	float sizex, sizey, sizez;
	float objx, objy, objz;
	float revflg;
	float diffuse;
	float highlight;
	float r, g, b;
};

struct objectlist{
	struct objectdef *obj;
	struct objectlist *next;
};

struct netdef{
	int netid;
	struct netdef *next;
};

struct andnetlist{
	struct netdef *anet;
	struct andnetlist *next;
};

struct ornetlist{
	int rpoid;
	struct netdef *onet;
	struct ornetlist *next;
};

struct sld{
	struct envdef *env;
	struct objectlist *objects;
	struct andnetlist *andnets;
	struct ornetlist *ornets;
};

struct sld *sldbody;

extern int yylineno;

void yyerror(const char *str){
	fprintf(stderr, "line(%d) parse error: %s\n", yylineno, str);
	existserr = 1;
}

int yywrap(){
	return 1;
}

void *malloc_noerr(size_t size){
	void *ret;
	
	ret = malloc(size);
	if(!ret){
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	
	return ret;
}

%}

%union{
	int ival;
	float fval;
	struct envdef *envd;
	struct objectdef *objd;
	struct objectlist *objl;
	struct netdef *netd;
	struct andnetlist *anetl;
	struct ornetlist *onetl;
	struct sld *sldb;
}

%token <ival> INT_VAL
%token <fval> FLOAT_VAL
%token EOL

%type <sldb> sldmain
%type <envd> envirom
%type <objd> object
%type <objl> objects
%type <anetl> andnlist
%type <netd> nd
%type <onetl> ornlist
%type <fval> intfloat


%%

sldmain:
	envirom objects andnlist ornlist {
		sldbody = (struct sld*)malloc_noerr(sizeof(struct sld));
		(sldbody->env) = $<envd>1;
		(sldbody->objects) = $<objl>2;
		(sldbody->andnets) = $<anetl>3;
		(sldbody->ornets) = $<onetl>4;
		$<sldb>$ = sldbody;
	}
;

envirom:
	intfloat intfloat intfloat intfloat intfloat INT_VAL intfloat intfloat intfloat {
		struct envdef *ev = (struct envdef*)malloc_noerr(sizeof(struct envdef));
		(ev->scrx) = $<fval>1;
		(ev->scry) = $<fval>2;
		(ev->scrz) = $<fval>3;
		(ev->seedeg1) = $<fval>4;
		(ev->seedeg2) = $<fval>5;
		(ev->lightnum) = $<ival>6;
		(ev->lightdeg1) = $<fval>7;
		(ev->lightdeg2) = $<fval>8;
		(ev->lightpow) = $<fval>9;
		$<envd>$ = ev;
	}
;

objects:
	object EOL {
		struct objectlist *ol = (struct objectlist*)malloc_noerr(sizeof(struct objectlist));
		(ol->obj) = $<objd>1;
		(ol->next) = NULL;
		$<objl>$ = ol;
	}
|	object objects {
		struct objectlist *ol = (struct objectlist*)malloc_noerr(sizeof(struct objectlist));
		(ol->obj) = $<objd>1;
		(ol->next) = $<objl>2;
		$<objl>$ = ol;
	}
;

object:
	INT_VAL INT_VAL INT_VAL INT_VAL intfloat intfloat intfloat intfloat intfloat intfloat intfloat intfloat intfloat intfloat intfloat intfloat {
		struct objectdef *ob = (struct objectdef*)malloc_noerr(sizeof(struct objectdef));
		(ob->texture) = $<ival>1;
		(ob->form) = $<ival>2;
		(ob->reftype) = $<ival>3;
		(ob->existrot) = $<ival>4;
		(ob->sizex) = $<fval>5;
		(ob->sizey) = $<fval>6;
		(ob->sizez) = $<fval>7;
		(ob->objx) = $<fval>8;
		(ob->objy) = $<fval>9;
		(ob->objz) = $<fval>10;
		(ob->revflg) = $<fval>11;
		(ob->diffuse) = $<fval>12;
		(ob->highlight) = $<fval>13;
		(ob->r) = $<fval>14;
		(ob->g) = $<fval>15;
		(ob->b) = $<fval>16;
		$<objd>$ = ob;
	}
;

nd:
	INT_VAL EOL {
		struct netdef *ndef = (struct netdef*)malloc_noerr(sizeof(struct netdef));
		(ndef->netid) = $<ival>1;
		(ndef->next) = NULL;
		$<netd>$ = ndef;
	}
|	INT_VAL nd {
		struct netdef *ndef = (struct netdef*)malloc_noerr(sizeof(struct netdef));
		(ndef->netid) = $<ival>1;
		(ndef->next) = $<netd>2;
		$<netd>$ = ndef;
	}
;

andnlist:
	nd EOL {
		struct andnetlist *al = (struct andnetlist*)malloc_noerr(sizeof(struct andnetlist));
		(al->anet) = $<netd>1;
		(al->next) = NULL;
		$<anetl>$ = al;
	}
|	nd andnlist {
		struct andnetlist *al = (struct andnetlist*)malloc_noerr(sizeof(struct andnetlist));
		(al->anet) = $<netd>1;
		(al->next) = $<anetl>2;
		$<anetl>$ = al;
	}
;

ornlist:
	INT_VAL nd EOL {
		struct ornetlist *ol = (struct ornetlist*)malloc_noerr(sizeof(struct ornetlist));
		(ol->rpoid) = $<ival>1;
		(ol->onet) = $<netd>2;
		(ol->next) = NULL;
		$<onetl>$ = ol;
	}
|	INT_VAL nd ornlist {
		struct ornetlist *ol = (struct ornetlist*)malloc_noerr(sizeof(struct ornetlist));
		(ol->rpoid) = $<ival>1;
		(ol->onet) = $<netd>2;
		(ol->next) = $<onetl>3;
		$<onetl>$ = ol;
	}
;

intfloat:
	INT_VAL {
		$<fval>$ = (float)($<ival>1);
	}
|	EOL {
		$<fval>$ = (float)(-1);
	}
|	FLOAT_VAL {
		$<fval>$ = $<fval>1;
	}
;


%%


void write_float(float f, FILE *fp){
	unsigned int u = *((unsigned int*)(&f));
	
	fprintf(fp, "%08x\n", u);
}

void write_int(int i, FILE *fp){
	unsigned int u = (unsigned int)i;
	
	fprintf(fp, "%08x\n", u);
}

void write_EOL(FILE *fp){
	unsigned int u = 0xffffffffU;
	
	fprintf(fp, "%08x\n", u);
}

int main(int argc, char *argv[]){
	extern FILE *yyin;
	char *outfilename;
	
	FILE *fp;
	
	if(sizeof(unsigned int) != 4){
		fprintf(stderr, "sizeof(unsigned int) must be 4\n");
		exit(1);
	}
	
	if(((0x80000001U >> 1) != 0x40000000U) || 
	   ((0x40000000U << 1) != 0x80000000U) || 
	   ((0x80000000U << 1) != 0x00000000U)){
		fprintf(stderr, "sll and srl needed\n");
		return 1;
	}
	
	if(((-1) >> 1) != -1){
		fprintf(stderr, "sra needed\n");
		return 1;
	}
	
	if(argc < 2){
		yyin = stdin;
		fp = NULL;
		outfilename = (char*)malloc(sizeof(char) * 11);
		if(!outfilename){
			fprintf(stderr, "malloc failed\n");
			exit(1);
		}
		strcpy(outfilename, "output.in");
	}
	else{
		fp = fopen(argv[1], "r");
		if(fp){
			char *s;
			
			yyin = fp;
			outfilename = (char*)malloc(sizeof(char) * (strlen(argv[1]) + 4));
			if(!outfilename){
				fprintf(stderr, "malloc failed\n");
				exit(1);
			}
			strcpy(outfilename, argv[1]);
			for(s = outfilename; (*s) && (*s != '.'); s++);
			*s = '.';
			*(s + 1) = 'i';
			*(s + 2) = 'n';
			*(s + 3) = '\0';
		}
		else{
			yyin = stdin;
			outfilename = (char*)malloc(sizeof(char) * 11);
			if(!outfilename){
				fprintf(stderr, "malloc failed\n");
				exit(1);
			}
			strcpy(outfilename, "output.in");
		}
	}
	
	yyparse();
	
	if(fp)
		fclose(fp);
	
	if(!existserr){
		if((fp = fopen(outfilename, "w"))){
			struct objectlist *ol;
			struct andnetlist *andl;
			struct ornetlist *orl;
			
			write_float((sldbody->env)->scrx, fp);
			write_float((sldbody->env)->scry, fp);
			write_float((sldbody->env)->scrz, fp);
			write_float((sldbody->env)->seedeg1, fp);
			write_float((sldbody->env)->seedeg2, fp);
			write_int((sldbody->env)->lightnum, fp);
			write_float((sldbody->env)->lightdeg1, fp);
			write_float((sldbody->env)->lightdeg2, fp);
			write_float((sldbody->env)->lightpow, fp);
			
			free(sldbody->env);
			
			for(ol = (sldbody->objects); ol; ){
				struct objectlist *olnxt = (ol->next);
				
				write_int((ol->obj)->texture, fp);
				write_int((ol->obj)->form, fp);
				write_int((ol->obj)->reftype, fp);
				write_int((ol->obj)->existrot, fp);
				write_float((ol->obj)->sizex, fp);
				write_float((ol->obj)->sizey, fp);
				write_float((ol->obj)->sizez, fp);
				write_float((ol->obj)->objx, fp);
				write_float((ol->obj)->objy, fp);
				write_float((ol->obj)->objz, fp);
				write_float((ol->obj)->revflg, fp);
				write_float((ol->obj)->diffuse, fp);
				write_float((ol->obj)->highlight, fp);
				write_float((ol->obj)->r, fp);
				write_float((ol->obj)->g, fp);
				write_float((ol->obj)->b, fp);
				
				free(ol->obj);
				free(ol);
				ol = olnxt;
			}
			
			write_EOL(fp);
			
			for(andl = (sldbody->andnets); andl; ){
				struct andnetlist *andlnxt = (andl->next);
				struct netdef *nd;
				
				for(nd = (andl->anet); nd; ){
					struct netdef *ndnxt = (nd->next);
					
					write_int(nd->netid, fp);
					
					free(nd);
					nd = ndnxt;
				}
				
				write_EOL(fp);
				
				free(andl);
				andl = andlnxt;
			}
			
			write_EOL(fp);
			
			for(orl = (sldbody->ornets); orl; ){
				struct ornetlist *orlnxt = (orl->next);
				struct netdef *nd;
				
				write_int(orl->rpoid, fp);
				
				for(nd = (orl->onet); nd; ){
					struct netdef *ndnxt = (nd->next);
					
					write_int(nd->netid, fp);
					
					free(nd);
					nd = ndnxt;
				}
				
				write_EOL(fp);
				
				free(orl);
				orl = orlnxt;
			}
			
			write_EOL(fp);
			
			free(sldbody);
			
			fclose(fp);
		}
		else{
			fprintf(stderr, "cannot create file %s\n", outfilename);
			exit(1);
		}
	}
	
	free(outfilename);
	
	return 0;
}
